#  - cluster.py - 
# Cluster mode support.
#
# This file is part of mp3togo
# 
# Convert audio files to play on a mp3 player
#
# (c) Simeon Veldstra 2006  <reallifesim@gmail.com>
#
# This software is free.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You may redistribute this program under the terms of the
# GNU General Public Licence version 2
# Available in this package or at http://www.fsf.org


"""Cluster mode.

	Run slave processes on multiple machines communicating by
	pipes over ssh and sharing data through NFS mounts."""

import os
import sys
import pty
import time
import fcntl
import select
import signal
import termios
import cPickle

import mp3togo.filelist as filelist
import mp3togo.track as track
import mp3togo.cache as cache
import mp3togo.conf as conf
import mp3togo.task as task
import mp3togo.pool as pool
import mp3togo


BEGIN = 'begin'
READY = 'ready'
INUSE = 'inuse'
EXIT  = 'exit'
 
MAXTRIES = 3

EOT = '\n*****\n\n'
EOR = '\n#####\n\n'
RS = '\n#**\n**#\n'

# Time to sleep between output messages
INTERVAL = 1.0


def slave(opts):
	"""Run as a slave process.

	The program was run with --cluster-slave. Establish
	communication with the master and do protocol.
	
	The message protocol:

	Enter:   Master --> options --> Slave-\              \
	                                      |              |
     /-------<-----------<----------------/              |
    |                                          EXIT      | BEGIN
	Slave --> ack --> Master-\                  ^        |
	                         |                  |        |
    /-------<-------<--------/                  |        /
    |                                           |        - READY
	|->Master --> [work_unit | quit] --> Slave--|        \
	|                                           |        |
	|  /--------------<---------------<---------/        | INUSE
	^  |                                                 |
	|  \ Slave --> % ... % ... % ... Done --> Master-\   /
	|                                                |
    \<----------------<---------------<--------------/   - READY	

	"""

	state = BEGIN
	sin = sys.stdin
	sout = sys.stdout
	cooked = None
	trk = None
	seen_pickle = True


	fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)   #Want

	def initialize(ropts):
		"""Integrate remote options into local opts"""
		st = ropts['pickle']
		opts._d = cPickle.loads(st)
		opts['verbosity'] = 0
		return "SLAVE: ready"

	def perform(rotps):
		"""Run the work unit and watch progress"""
		job_start = time.time()
		if not rotps.has_key('file'):
			raise conf.ErrorClusterProtocol
		if opts['usecache']:
			fc = cache.Cache(opts['usecache'], opts)
		else:
			fc = None
		trk = track.Track(rotps['file'], opts, fc)
		pl = pool.Pool(opts)
		if pl.add_track(trk):
			tsk = trk.getfirst()
			while tsk:
				tsk.run()
				while tsk.status() == task.RUNNING:
					yield "Progress: %s %s\n" % (tsk.name, str(tsk.output()))
				tsk.wait()
				if tsk.status() != task.DONE:
					name = tsk.name
					while tsk:
						tsk.undo()
						tsk = tsk.prev()
					yield "FAIL: %s\n" % name
					raise StopIteration
				tsk = tsk.next()
		else:
			yield "FAIL: Insufficient space\n" 
			raise StopIteration
		yield "DONE: %f\n" % (time.time() - job_start)
		raise StopIteration


	# The state machine
	while state != EXIT:
		if state == INUSE:
			raise conf.ErrorClusterProtocol
		# Read from master
		rdata = ''
		while rdata[-len(EOT):] != EOT:
			rfd, wfd, xfd = select.select([sys.stdin.fileno()], 
			                              [], [], 15) 
			if not rfd:
				sout.write("No input within fifteen seconds, ")
				sout.write("can't wait around forever.")
				state = EXIT
				break
			rdata += sin.read() 
			if rdata == "SLAVE EXIT":
				state = EXIT
				break
		sout.flush()
		sin.flush()
		rdata = rdata[:-len(EOT)]
		rdata = rdata.replace('\r', '')
		ropts = {}
		for line in rdata.split(RS):
			if '=' in line:
				key, value = line.split('=', 1)
				ropts[key] = value[:-1] # \n
	
		# Write to master
		if state == BEGIN:
			msg = initialize(ropts)
			if msg:
				state = READY
			else:
				state = EXIT
				msg = "Did not understand request."
			msg += EOR
			sout.write(msg)
			sout.flush()
			sin.flush()
		elif state == READY:
			state = INUSE
			gen = perform(ropts)
			while state == INUSE:
				try:
					msg = gen.next()
					time.sleep(INTERVAL)
				except StopIteration:
					state = READY
					msg = EOR
				sout.write(msg)
				sout.flush()
				sin.flush()

	# END REMOTE END

	# START MASTER END

class Boss:
	"""Manage one slave"""

	def __init__(self, host, files, failures, opts, r_schema=""):
		"""Manage a remote host. use r_schema='ssh %h ' for remote hosts
		Files is a FileList of files to process the list should be
		shared with the cooperating bosses"""
		self.host = host
		self.files = files
		self.fails = failures
		self.opts = opts
		self.current = None
		self.pid, self.fd1 = pty.fork()
		if self.pid == 0:
			args = r_schema.replace('%h', host)
			args += " mp3togo --cluster-slave True"
			args = args.split()
			os.execvp(args[0], args)
		else:
			self.fd2 = os.dup(self.fd1)
			self.rf = os.fdopen(self.fd1, 'r', 0)
			fcntl.fcntl(self.fd1, fcntl.F_SETFL, os.O_NONBLOCK)
			self.wf = os.fdopen(self.fd2, 'w', 0)
			fcntl.fcntl(self.fd2, fcntl.F_SETFL, os.O_NONBLOCK)

	def __del__(self):
		self.rf.flush()
		self.wf.flush()
		self.rf.close()
		self.wf.close()
		try:
			os.waitpid(self.pid, 0)
		except OSError:
			pass

	def poll(self):
		start = 0
		last_action = ''
		optd = self.opts._d.copy()
		optd['arg_files'] = []
		st = "pickle=" + cPickle.dumps(optd) + '\n'
		self.wf.write(st)
		self.wf.flush()
		self.wf.write(EOT)
		self.wf.flush()
		errors = 0
		while 1:
			a, b, c = select.select([self.fd1, self.fd2], [], [], 60)
			if a:
				try:
					inp = self.rf.read()
				except IOError:
					errors += 1
					if errors > 25:
						if self.current:
							self.files.push(self.current, front=True)
						raise StopIteration
					yield ('Not Ready', 'Not Ready', 0)
					continue
				self.rf.flush()
				for line in inp.split('\n'):
					line = line.replace('\r', '')
					def next():
						try:
							self.current = self.files.pop(0)
						except:
							self.wf.write('SLAVE EXIT')
							os.waitpid(self.pid, 0)
							raise StopIteration
						if self.fails.has_key(self.current) and (
						            self.host in self.fails[self.current]
									or len(self.fails[self.current]) > MAXTRIES):
							oldcurrent = self.current
							try:
								next()
							finally:
								self.files.push(oldcurrent, front=True)
							return
						self.wf.write('file=%s\n' % self.current)
						self.wf.write(EOT)
						start = time.time()
						percent = 0.0
					if EOR.replace('\n', '') == line:
						if self.current:
							yield (self.current, 'Finished', 
							       float(time.time() - start))
						next()
						break
					elif line.startswith('Progress:'):
						action = ' '.join(line.split()[1:-1]) 
						if not action:
							action = last_action
						if action != last_action:
							percent = 0.0
						last_action = action
						try:
							percent = float(line.split()[-1])
						except ValueError:
							pass
						yield (self.current, action, percent)
					elif line.startswith('FAIL:'):
						if "Insufficient space" in line:
							yield (self.current, 'Insufficient space', 0.0)
							raise StopIteration
						else:
							yield (self.current, 'Subtask Failed', 0.0)
							if self.fails.has_key(self.host):
								self.fails[self.current].append(self.host)
							else:
								self.fails[self.current] = [self.host]
								oldcurrent = self.current
							try:
								next()
							finally:
								self.files.push(oldcurrent, front=True)
							break
			else:
				#Timed out
				self.kill()
				raise StopIteration

	def kill(self):
		"""Kill the child
		This kills the ssh connection, it is 
		unable to clean up the remote machine."""
		os.kill(self.pid, signal.SIGKILL)
		os.waitpid(self.pid, 0)



def slavedriver(playlist, opts, cooked=None):
	"""A simple text mode cluster controller.
	
	Function signature for running from main.py"""

	if opts['verbosity']:
		def p(s):
			sys.stdout.write(s)
	else:
		p = lambda x: None
	L = opts.log

	p("mp3togo version %s\n" % mp3togo.version)
	p("Starting Cluster mode\n\n")

	slaves = opts['cluster']
	slaves = slaves.split(',')
	master = None

	p("Connecting to cluster nodes:\n")
	fails = {} # {'badfilename': [host,...]}
	if opts['clusternolocal']:
		ring = []
	else:
		p("  localhost")
		master = Boss('localhost', playlist, fails, opts, r_schema='')
		ring = [('localhost', master.poll(), master)]
	for slave in slaves:
		p("  " + slave)
		theman = Boss(slave, playlist, fails, opts, r_schema="ssh %h ")
		tr = (slave, theman.poll(), theman)
		ring.append(tr)
	

	p("\n\n")
	files = dict(map(lambda x: (x, ''), slaves))
	if master:
		files[master.host] = ''
	while ring:
		for slave in ring[:]:
			try:
				tup = slave[1].next()
				#print tup
			except StopIteration:
				L(1, "Host %s finished.\n" % slave[0])
				del ring[ring.index(slave)]
				break
			if tup[1] == 'Subtask Failed':
				L(1, "Task failed on host: %s\n  %s\n" % (
				      slave[0], tup[0]))
			if tup[0] != files[slave[0]]:
				files[slave[0]] = tup[0]
				L(1, "Host %s at %s starting file: \n   %s\n" % (
				      slave[0], time.ctime(), tup[0]))
		if ring:
			time.sleep(0.5)
	print "All workers exited"
	
# Test that all files are actually done, if not, loop.

	


