#  - task.py - 
# This file is part of mp3togo

# Convert audio files to play on a mp3 player
# Manage a sub process.
#
# (c) Simeon Veldstra 2006  <reallifesim@gmail.com>
#
# This software is free.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You may redistribute this program under the terms of the
# GNU General Public Licence version 2
# Available in this package or at http://www.fsf.org


import os
import sys
import time
import select
import pty
import fcntl
import signal
import threading
import types

import mp3togo.conf as setup


READY = 'ready'
RUNNING = 'running'
PAUSED = 'paused'
STOPPED = 'stopped'
FAILED = 'failed'
DONE = 'done'


class SimpleTask:
	"""Run a callable and save its output"""
	
	def __init__(self, parent, action, fltr=None, reverse=None, tmpsize=0, outsize=0, name=''):
		if not action:
			raise TypeError
		if str(self.__class__).endswith('SimpleTask'):
			if not callable(action):
				raise TypeError
		for func in (fltr, reverse):
			if func:
				if not callable(func):
					raise TypeError

		self.name = name
		self.tmpsize = tmpsize
		self.outsize = outsize
		self._parent = parent
		self._action = action
		self._filter = fltr
		self._reverse = reverse
		# Hold runlock while accessing self._status
		self._runlock = threading.Lock()
		# Hold outlock while accessing self._output
		self._outlock = threading.Lock()
		self._status = READY
		self._output = None
		self._start_time = 0
		self._finish_time = 0

	def run(self):
		"""Run the callable"""
		self.check_sibling()
		self._runlock.acquire()
		if self._status != READY:
			self._runlock.release()
			return False
		self._status = RUNNING
		self._start_time = time.time()

		try:
			out = self._action()
			if self._filter:
				out = self._filter(out)
			self._outlock.acquire()
			self._output = out
			self._outlock.release()
		except:
			# if verbosity is high enough, print something about the exception
			self._status = FAILED
			self._runlock.release()
			return True
		self._status = DONE
		self._finish_time = time.time()
		self._runlock.release()
		return True

	def pause(self):
		return False

	def stop(self):
		return False

	def undo(self):
		"""Undo the action, if possible.

		Runs the reverse function passed in.
		reverse should return a True value on success or
		False in the event of a failure. undo returns None 
		if no reverse function was passed in.

		If reverse succeeds, the task object should be in a
		consistent READY state and run() can start again at the 
		top. If it fails or throws an exception, the state will
		remain as it is.

		"""
		self._runlock.acquire()
		if self._status not in (DONE, FAILED, STOPPED):
			self._runlock.release()
			raise setup.TaskNotReadyError
		ret = None
		if self._reverse:
			try:
				ret = self._reverse()
			except:
				self._runlock.release()
				return False
		if ret:
			self._status = READY
		self._runlock.release()
		return ret

	def status(self):
		self._runlock.acquire()
		ret = self._status
		self._runlock.release()
		return ret

	def wait(self):
		pass

	def wait_unpause(self):
		pass

	def space_req(self):
		return (self.tmpsize, self.outsize)

	def output(self):
		self._outlock.acquire()
		out = self._output
		self._outlock.release()
		return out

	def elapsed_time(self):
		if not self._finish_time:
			return False
		return self._finish_time - self._start_time

	def check_sibling(self):
		bro = self.prev()
		if bro and bro.status() != DONE:
			raise setup.TaskNotReadyError

	def next(self):
		"""Return the next task in the queue."""
		try:
			queue = self._parent.tasks()
		except: 
			return None
		qi = list(queue).index(self)
		if qi + 1 < len(queue):
			return queue[qi + 1]

	def prev(self):
		"""Return the previous task in the queue."""
		try:
			queue = self._parent.tasks()
		except: 
			return None
		qi = list(queue).index(self)
		if qi - 1  >= 0:
			return queue[qi - 1]

	
class Task(SimpleTask):
	"""Run a command in a subprocess and monitor its progress"""

	def __init__(self, parent, action, filter=None, reverse=None, tmpsize=0, outsize=0, name=''):
		if type(action) not in (list, tuple):
			raise TypeError
		for element in action:
			if type(element) not in types.StringTypes:
				raise TypeError
		self._paused = threading.Lock()
		self._pid = None
		self._eater = None
		SimpleTask.__init__(self, parent, action, filter, reverse, tmpsize, outsize, name)

	def run(self):
		"""Run the command in a sub process
		
		This will probably only work on Linux.
		
		"""

		self.check_sibling()
		self._runlock.acquire()
		self._start_time = time.time()
		if self._status == PAUSED:
			self._paused.release()
			self._status = RUNNING
			self._runlock.release()
			return True
		if self._status != READY:
			self._runlock.release()
			return False
		self._status = RUNNING

		fp = None

		def eatout():
			while not fp:
				pass
			while self._status in (RUNNING, PAUSED):
				wpid, status = os.waitpid(self._pid, os.WNOHANG)
				if wpid == pid:
					# Child exited
					self._runlock.acquire()
					if self._paused.locked():
						self._paused.release()
					if os.WIFSIGNALED(status):
						#if os.WSTOPSIG(status) == 0:
						#	self._status = DONE
						#else:
						self._status = FAILED
					elif os.WEXITSTATUS(status) == 0:
						self._status = DONE
					else:
						self._status = FAILED
					self._runlock.release()
					break
				if self._paused.locked():
					continue
				rfd, wfd, xfd = select.select([master_fd], [], [], 5)
				if not rfd:
					break
				try:
					if master_fd in rfd:
						out = fp.read()
				except IOError:
					break
				if self._filter:
					try:
						out = self._filter(out)
					except:
						out = ''
				self._outlock.acquire()
				self._output = out
				self._outlock.release()
			# Child still running, no more output: block on child's exit
			if self._status in (RUNNING, PAUSED):
				self._runlock.acquire()
				if self._paused.locked():
					self._paused.release()
				try:
					wpid, status = os.waitpid(pid, 0) 
					if os.WIFSIGNALED(status):
						if os.WTERMSIG(status) == signal.SIGHUP:
							self._status = FAILED
						else:
							self._status = FAILED
					elif os.WEXITSTATUS(status) == 0:
						self._status = DONE
					else:
						self._status = FAILED
				except OSError:
					self._status = FAILED
				self._runlock.release()
			self._finish_time = time.time()

		self._eater = threading.Thread(None, eatout)
		self._eater.start()

		pid, master_fd = pty.fork()
		if pid == 0:
			time.sleep(0.0125)
			try:
				os.execvp(self._action[0], self._action)
			except:
				os._exit(1)
		self._pid = pid
		fp = os.fdopen(master_fd, 'r', 0)
		if os.uname()[0] == "Linux":
			fcntl.fcntl(master_fd, fcntl.F_SETFL, os.O_NONBLOCK)
		time.sleep(0.0125)
		self._runlock.release()
		return True
		
	def pause(self):
		self._runlock.acquire()
		if self._status == RUNNING:
			self._paused.acquire()
			self._status = PAUSED
			self._runlock.release()
			return True
		else:
			self._runlock.release()
			return False
	
	def stop(self):
		"""Stop the running process"""
		self._runlock.acquire()
		if self._status == RUNNING:
			try:
				os.kill(self._pid, signal.SIGKILL)
				self._status = STOPPED
				self._runlock.release()
				self._eater.join()
				return True
			except OSError:
				pass
		self._runlock.release()
		return False

	def wait(self):
		if self._eater:
			self._eater.join()

	def wait_unpause(self):
		self._paused.acquire()
		self._paused.release()

